/*

##################################
This program is enhanced by: 
    PUT IN YOUR NAME!!!
##################################


This program manages And/Or-Sequences of <Strings>.  Such sequences
consist of <Element> objects. There are three kinds of <Elements>:

  1. <String_Element> -- which contains a single <String>.

  2. <And_Element> -- which contains a list of elements that are
     concatenated in the given order.

  3. <Or_Element> -- which contains a set of elements that are used
     alternatively, exactly one at a time.

The <And_Element> and the <Or_Element> objects can be nested
arbitrarily.  The lowest level (leafs) of any object tree formed this
way always contains <String_Element> objects.  <Or_Elements> result in
variants.  Where <Or_Elements> meet, the cross product of their
variants is formed.

Here is the class hierarchy:

  abstract class Element
     class String_Element
     class And_Element
     class Or_Element

  abstract class Element_Action
     class Depth

The main function generates an And/Or sequence and prints it.

Note, do not worry about memory leaks in this code.
*/

// Forward declarations.
class Element;
class String_Element;
class And_Element;
class Or_Element;

// These are necessary to provide String and Queue abstractions.

#include "ace/SString.h"
#include "ace/Containers.h"

// Some simple typedefs.
typedef ACE_CString String;
typedef ACE_Unbounded_Queue<Element *> Element_Queue;
typedef ACE_Unbounded_Queue_Iterator<Element *> Element_Queue_Iterator;

class Element_Action 
  // = TITLE
  //
  //   Interface for operations on Element structures that must handle
  //   the parts differently, depending on their type.
{
public:
  // Perform the action for <String_Elements>.
  virtual void string_action (String_Element &e) = 0;

  // Perform the action for <And_Elements>.
  virtual void and_action (And_Element &e) = 0;

  // Perform the action for <Or_Elements>.
  virtual void or_action (Or_Element &e) = 0;
};

class Element
  // = TITLE
  //
  //    Interface of the 'Element' classes, i.e., String_Element,
  //    And_Element, Or_Element.
  // 
{
public:
  // Input operation. Adds another element to this element.
  virtual void add (Element *e) = 0;

  // Output operation.  Prints the whole <Element> on <cout>.  Each
  // variant is printed on a separate line.
  virtual String to_string (void) = 0;

  // Output operation.  Prints the whole <Element> on <cout>.  Each
  // variant is printed on a separate line.
  virtual void print (void) 
    {
      Element_Queue eq = this->variants ();
      Element **e = 0;

      // Iterate through all the <Elements> and print each on a
      // separate line.
      for (Element_Queue_Iterator iter (eq);
           iter.next (e) != 0;
           iter.advance ())
        // Print out the char * representation of the string.
        cout << (*e)->to_string () << endl;
    }

  // Representation operation. Returns a <Element_Queue> in which each
  // element contains the <String> representation of one variant of
  // the And/Or sequence.  Each variant appears exactly once in the
  // <Element_Queue>.
  virtual Element_Queue variants (void) = 0;

  // Branching operation. Calls that one operation from
  // <Element_Action> that is responsible for (i.e., corresponds to)
  // the present subclass.
  virtual void perform (Element_Action &a) = 0;
};

class String_Element : public Element
  // = TITLE
  //
  //     Element class that contains exactly one String.
  //
{
public:
  // Generates a String_Element with empty contents.
  String_Element (void) 
    {
    }

  // Generates a String_Element with the contents given as argument.
  String_Element (const String &s) 
    : contents_ (s)
    {
    }

  // Input operation. Appends <e> to current String contents if <e> is
  // a <String_Element>.  We assume that <e> is a <String_Element>.
  virtual void add (Element *e)
    {
      // Perform a cast (oh what I wouldn't give for RTTI... ;-));
      String_Element &s = (String_Element &) e;

      // += is the append operator.
      this->contents_ += s.contents_;
    }

  // Representation operation. Returns a coded representation of the
  // element as a string of the form <contents_>.
  virtual String to_string (void) 
    {
      return this->contents_;
    }
  
  // Returns a <Element_Queue> with exactly one element: the String
  // that the <String_Element> contains.
  virtual Element_Queue variants (void)
    {
      Element_Queue sa;
      sa.enqueue_tail (this);
      return sa;
    }
  
  // Calls the string_action of the Element_Action given as argument. 

  virtual void perform (Element_Action &a)
    {
      a.string_action (*this);
    }

private:
  // Contents of the String_Element.
  String contents_;     
};

class And_Element : public Element 
  // = TITLE  
  //
  //     And sequence of elements.  That means the elements must be
  //     concatenated. 
  //
{
public:
  // Generates an <And_Element> without any contents.
  And_Element (void) 
    {
    }

  // Input operation. Adds <e> as new element to current sequence of
  // elements.
  virtual void add (Element *e)
    {
      this->elements_.enqueue_tail (e);
    }

  // Representation operation.  Returns a coded representation of the
  // Element as a String of the form "AND ( el1 & el2 & el3 )".
  virtual String to_string (void)
    {
      String b ("AND ( ");

      Element **e = 0;

      // Iterate through the set of <elements_> and convert them into
      // a String.

      Element_Queue_Iterator iter (this->elements_);

      while (iter.next (e) != 0)
	{
	  b += (*e)->to_string ();

	  // Check to see if we're at the end of the iteration.
	  if (iter.advance () != 0)
	    b += String ("& ");
	}

      b += String (") ");

      return b;
    }
  
  // Returns cross product of all variants of all elements in the
  // <And_Element>.  The product is formed by concatenating the
  // Strings of the variants.

  virtual Element_Queue variants (void)
    {
      Element_Queue result;
      Element **e = 0;

      // Iterate over all the elements.

      for (Element_Queue_Iterator iter (this->elements_);
           iter.next (e) != 0;
           iter.advance ())
        {
          Element_Queue nextpart ((*e)->variants ());

          if (nextpart.size () == 0)
            continue; // Nothing to add to result for this element.
          if (result.size () == 0) 
            {
              result = nextpart;
              continue;  // Nothing to combine with yet.
            }

          // Store <result> in <oldreset> and reset <result> to be
          // empty in preparation for the cross product calculation.
          Element_Queue oldresult = result;
          result.reset ();

          size_t old_n = oldresult.size ();
          size_t next_n = nextpart.size ();

          // Generate cross product (by String concatenation):
	  // 
          // result := oldresult x nextpart

          for (size_t old_i = 0; 
               old_i < old_n;
               old_i++) 
            {
              Element **ith_element;
              // Get the "ith" element in <oldresult>.
              oldresult.get (ith_element, old_i);

              String current_old = (*ith_element)->to_string ();

              for (size_t next_i = 0; 
                   next_i < next_n; 
                   next_i++) 
                {
                  // Find the "ith" element in <nextpart>.
                  nextpart.get (ith_element, next_i);

                  String current_next = (*ith_element)->to_string ();

                  // Concatenate the strings.
		  String_Element *concat =
		    new String_Element (current_old + current_next);

		  // Insert the concatenated strings at the end of the
		  // result.
                  result.enqueue_tail (concat);
                }
            }
        }

      return result;
    }
  
  // Calls the string_action of the Element_Action given as argument. 

  virtual void perform (Element_Action &a)
    {
      a.and_action (*this);
    }

  // @@ private:
  // Array of the elements forming the <And_Element>.
  Element_Queue elements_;   
};

class Or_Element : public Element
  // = TITLE
  //
  //    Or set of elements.
  //    (That means the elements are alternatives.)
  //
{
public:

  // Generates an Or_Element without Contents.
  Or_Element (void) 
    {
    }

  // Input operation. Adds e as new element to current set of
  // alternative elements. 

  virtual void add (Element *e)
    {
      this->elements_.enqueue_tail (e);
    }

  // Representation operation. Returns a coded representation of the
  // Element as a String of the form "OR (el1 | el2 | el3 )".

  virtual String to_string (void)
    {
      String b ("OR ( ");

      Element **e = 0;

      // Iterate through the set of <elements_> and convert them into
      // a String.

      Element_Queue_Iterator iter (this->elements_);

      while (iter.next (e) != 0)
	{
	  b += (*e)->to_string ();

	  // Advance the iterator 
	  if (iter.advance () != 0)
	    b += String ("| ");
	}

      b += String (") ");

      return b;
    }
  
  // Returns the union of all variants of all <elements_> in the
  // <Or_Element>.

  virtual Element_Queue variants (void)
    {
      Element_Queue result;

      Element **oe = 0;

      for (Element_Queue_Iterator outer_iter (this->elements_);
           outer_iter.next (oe) != 0;
           outer_iter.advance ())
        {
          Element_Queue nextpart = (*oe)->variants ();

          Element **ie = 0;

          for (Element_Queue_Iterator inner_iter (nextpart);
               inner_iter.next (ie) != 0;
               inner_iter.advance ())
            result.enqueue_tail (*ie);
        }
      
      return result;
    }
  
  // Calls the string_action of the Element_Action given as argument.

  virtual void perform (Element_Action &a)
    {
      a.or_action (*this);
    }

  // @@ private:
  Element_Queue elements_; // vector of the elements forming the Or_Element
};

class Depth : public Element_Action 
  // = TITLE
  //
  //  Class for computing maximum depths of the different kinds of nodes
  //  in an And/Or Element tree.  
{
public:  
  // Walks through Element e to initialize itself.  Components
  // max_depth, max_and_depth, and max_or_depth can be queried afterwards.

  Depth (Element *e)
    : max_depth_ (-1),
      max_and_depth_ (-1),
      max_or_depth_ (-1),
      depth_ (0)
    {
      e->perform (*this);
    }
  
  virtual void string_action (String_Element &)
    {
      if (this->max_depth_ < this->depth_)
        this->max_depth_ = this->depth_;
    }
  
  virtual void and_action (And_Element &u)
    {
      if (this->max_and_depth_ < this->depth_) 
        {
          this->max_and_depth_ = this->depth_;

          if (this->max_depth_ < this->depth_)
            this->max_depth_ = this->depth_;
        }

      this->iterate (u.elements_);
    }
  
  virtual void or_action (Or_Element &o)
    {
      if (this->max_or_depth_ < this->depth_) 
        {
          this->max_or_depth_ = this->depth_;

          if (this->max_depth_ < this->depth_)
            this->max_depth_ = this->depth_;
        }

      this->iterate (o.elements_);
    }
  
  // = These should be private ;-)
  int max_depth_;  // Depth of deepest Element
  int max_and_depth_;  // Depth of deepest And_Element
  int max_or_depth_;  // Depth of deepest Or_Element

private:
  // Handle all elements contained in an <And_Element> or an
  // <Or_Element>.

  void iterate (Element_Queue &elems)
    {
      this->depth_++;  // these elements are one level deeper

      Element **e;

      for (Element_Queue_Iterator iter (elems);
           iter.next (e) != 0;
           iter.advance ())
        (*e)->perform (*this);

      this->depth_--;
    }

  int depth_;  // current depth (root == 0)  
};

// main program. generates an And/Or sequence and prints it.

int 
main (int, char *[])
{
  And_Element *u = new And_Element;
  Or_Element *modal = new Or_Element;
  Or_Element *verb = new Or_Element;

  u->add (new String_Element ("Who "));
  modal->add (new String_Element ("can "));
  modal->add (new String_Element ("will "));
  u->add (modal);
  verb->add (new String_Element ("think "));
  verb->add (new String_Element ("act "));
  u->add (verb);
  u->add (new String_Element ("after all"));
  u->add (new String_Element ("?"));

  u->print ();

  cout << u->to_string ();

  Depth t (u);
  cout << endl
       << "  Depth = " << t.max_depth_
       << "  AndDepth = " << t.max_and_depth_
       << "  OrDepth = " << t.max_or_depth_ << endl;
  return 0;
}
