/*

##################################
This program is enhanced by: 
    PUT IN YOUR NAME!!!
##################################

This program manages sets of tuples.

A tuple consists of several fields, e.g., last name, first name,
telephone number.  There are several such tuple types and for each of
them there is a tuple set type.

Furthermore, the program contains classes for displaying tuple sets of
a particular tuple type.  Currently only the tuple type
Name_Number_Tuple is fully supported.

This is the class hierarchy:

  abstract class Tuple
     class Name_Number_Tuple

  class Tupleset

  abstract class Tuple_Display
     class Name_Number_Tuple_Display_1
     abstract class Tuple_Display_A
        class Name_Number_Tuple_Display_2

Note, do not worry about memory leaks in this code.

#M *** DESIGN PATTERN: ***
#M
#M <Tuple_Display> is the abstract Superclass of **Observer**
#M for the data structure <Tupleset>.
#M
#M <Tuple_Display_A> is the abstract Superclass of a number of such 
#M observers that differ with respect to selection, ordering, and 
#M formatting of the tuples to be shown.  A **Template Method** 
#M is used to vary these aspects.  */

// Forward declarations.
class Tuple;
class Tuple_Display;
class Tupleset;

// These are necessary to provide String and Queue abstractions.

#include "ace/SString.h"
#include "ace/Containers.h"

// Some typedefs.
typedef ACE_CString String;
typedef ACE_Unbounded_Queue<Tuple *> Tuple_Queue;
typedef ACE_Unbounded_Queue_Iterator<Tuple *> Tuple_Queue_Iterator;
typedef ACE_Unbounded_Queue<Tuple_Display *> Tuple_Display_Table;
typedef ACE_Unbounded_Queue_Iterator<Tuple_Display *> Tuple_Display_Table_Iterator;
typedef ACE_Unbounded_Queue<String> Tuple_String_Vector;
typedef ACE_Unbounded_Queue_Iterator<String> Tuple_String_Vector_Iterator;

class Tuple_Display
  // = TITLE
  //
  // Tuple_Display shows multiple Tuples on the output stream.  Each
  // subclass defines its own presentation style.  Tuple_Displays
  // guarantee that they do not modify the Tuple objects (so that it
  // is possible to commit originals to the Tuple_Display's custody
  // instead of creating copies).  A program may create a number of
  // Tuple_Display objects and give the same Tuple objects to each of
  // them.
{
public:
  Tuple_Display (const String &name)
    : display_name_ (name)
    {
    }
     
  // Is called by the user of the <Tuple_Display> to announce that a
  // new <Tuple> should be added to the ones already displayed.
  virtual void new_tuple (Tuple *t) = 0;

  // Is called by the user of the <Tuple_Display> to announce that
  // multiple new <Tuples> shall be added to the ones already
  // displayed.  The caller guarantees that all objects in the
  // <Tuple_Queue> are <Tuple> objects.  <new_tuples> does not change
  // <tq>.
  virtual void new_tuples (Tuple_Queue &tq) = 0;

protected:
  // Queue of tuples.
  Tuple_Queue tuple_queue_;

  // Name of the display;
  String display_name_;
};

class Tupleset 
  // = TITLE
  //
  //    Manages a set of <Tuples> and allows individual Tuples to be
  //    added to this <Set>.  Also manages a set of <Tuple_Displays>,
  //    which are notified when a <Tuple> is added.
{
public:
  // Called by a <Tuple> generator to bring a new <Tuple> into the
  // <Tupleset>.
  void new_tuple (Tuple *t)
    {
      // Store <t> at the end of the queue.
      tuple_queue_.enqueue_tail (t);

      Tuple_Display **td = 0;

      // Inform each <Tuple_Display> that a new tuple has arrived.
      for (Tuple_Display_Table_Iterator iter (this->tuple_display_table_);
	   iter.next (td) != 0;
	   iter.advance ())
	(*td)->new_tuple (t);
    }

  // Add a new display.
  void new_display (Tuple_Display *td)
    {
      // Store <td> at the end of the queue.
      this->tuple_display_table_.enqueue_tail (td);

      // Inform the new display about all the existing <Tuples>.
      td->new_tuples (this->tuple_queue_);
    }

private:
  // Registered Displays.
  Tuple_Display_Table tuple_display_table_;

  // Stores the Tuples.
  Tuple_Queue tuple_queue_;
};

class Tuple 
  // = TITLE
  //
  //    Common interface of all tuple types.
{
public:
  // Reads a tuple from <cin> and stores it in the <Tupleset>.
  virtual void get_tuple (Tupleset &ts, istream &is) = 0;
};

class Name_Number_Tuple : public Tuple
  // = TITLE
  //
  //    An implementation of Tuple for names and telephone numbers. 
{
public:
  // = Initialization methods.
  Name_Number_Tuple (void)
    {
    }

  Name_Number_Tuple (const String &lastname,
		     const String &firstname,
		     const String &telephone)
    : lastname_ (lastname), 
      firstname_ (firstname),
      telephone_ (telephone)
    {
    }

  // = Accessor and mutator methods.

  String lastname (void) { return this->lastname_; }
  String firstname (void) { return this->firstname_; }
  String telephone (void) { return this->telephone_; }

  void lastname (const String &l) { this->lastname_ = l; }
  void firstname (const String &f) { this->firstname_ = f; }
  void telephone (const String &t) { this->telephone_ = t; }
  
  // Reads a <Tuple>. The exact strategy used is encapsulated in
  // <get_tuple>.  <get_tuple> guarantees that the <Tuple> will be
  // passed to <Tupleset> s after reading and that its fields contain
  // the correct values.

  void get_tuple (Tupleset &ts, istream &is)
    {
      char lastname[BUFSIZ];
      char firstname[BUFSIZ];
      char telephone[BUFSIZ];

      // Runs the event loop that gets the input tuple from the input
      // stream and updates the display.

      for (;;)
	{
	  cout << "enter lastname, firstname, phonenumber" << endl
	       << "(separated by spaces)" << endl;

	  is >> lastname >> firstname >> telephone;

	  if (is)
	    {
	      Name_Number_Tuple *nntp = 
		new Name_Number_Tuple (lastname,
				       firstname,
				       telephone);

	      // Insert the value of this tuple into the <Tupleset>.
	      ts.new_tuple (nntp);
	    }
	  else
	    break;
	}
    }

  String lastname_;      // Name of Person
  String firstname_;   // Firstname of Person
  String telephone_;   // Telephonenumber of Person
};

class Name_Number_Tuple_Display_1 : public Tuple_Display
  // = TITLE
  //
  // Displays <Name_Number_Tuple> objects in the order in which they
  // are delivered, showing all their components in a simple format.
{
public:
  Name_Number_Tuple_Display_1 (const String &name)
    : Tuple_Display (name)
    {
    }

  virtual void new_tuple (Tuple *t)
    {
      Name_Number_Tuple *nnt = (Name_Number_Tuple *) t;

      // Display this Tuple on output stream.
      String s;
      s += nnt->firstname ();
      s += String (" ");
      s += nnt->lastname ();
      s += String (", ");
      s += nnt->telephone ();
      s += String ("\n");

      this->tuple_string_vector_.enqueue_tail (s);
      this->display ();
    }

  virtual void new_tuples (Tuple_Queue &tq)
    {
      Tuple **t = 0;

      // Display all the tuples in the <Tuple_Queue>.
      for (Tuple_Queue_Iterator iter (tq);
	   iter.next (t) != 0;
	   iter.advance ())
	this->new_tuple (*t);
    }

  // Displays the contents of <tuple_string_vector_> on the output
  // stream.

  void display (void)
    {
      if (this->tuple_string_vector_.size () > 0)
	{
	  cout << "===========" << endl
	       << "Displaying " << this->display_name_ << endl;

	  String *s = 0;
	  size_t count = 0;

	  for (Tuple_String_Vector_Iterator iter (this->tuple_string_vector_);
	       iter.next (s) != 0;
	       iter.advance ())
	    cout << count++ << ": " << *s;

	  cout << "===========" << endl << endl;
	}
    }

private:
  // <tuple_string_vector_> contains the display presentation of the
  // <Tuples>.  The contents of <tuple_string_vector_> are written
  // into the output stream as is.
  Tuple_String_Vector tuple_string_vector_; 
};

class Tuple_Display_A : public Tuple_Display 
  // = TITLE
  //
  //    Tuple_Display_A displays Tuplesets in a style that can be
  //    adapted to many different purposes by inserting an appropriate
  //    selection method (reject Tuples completely), comparison method
  //    (define a Tuple order) and formatting method (select Tuple
  //    components and format them).
{
public:
  Tuple_Display_A (const String &name)
    : Tuple_Display (name)
    {
    }

  // Returns desired representation of Tuple a as a String.  The
  // String may contain zero, one, or more newline characters.

  virtual String format (Tuple *a) = 0;

  // Returns true, if the Tuple should be displayed and false otherwise.
  virtual int select (Tuple *a) = 0;

  // Returns true, if Tuple a should be displayed before Tuple b and
  // false, if Tuple b should be displayed before Tuple a.
  virtual int less_than (Tuple *a, Tuple *b) = 0;

  // Implements adding a new <Tuple>.  First <select> is used to test,
  // whether the <Tuple> should be added at all, then <merge_in> moves
  // it to the right place in the presentation using <less_than> and
  // <format> converts it into a String of the desired display format.
  //
  // #M      *** DESIGN PATTERN: ***
  // #M      new_tuple () together with its auxiliary method merge_in () forms a
  // #M      **Template Method**. The empty spots that are filled in subclasses
  // #M      are the methods select (), format (), and less_than ().

  virtual void new_tuple (Tuple *name_number)
    {
      if (this->select (name_number) == 0)
        return; // This Tuple won't be displayed

      // Puts name_number into tuple_queue_ and tuple_string_vector_
      this->merge_in (name_number);  
      this->display ();
    }

  // Analog to <new_tuple>, but uses a loop for adding several
  // <Tuples> at once. The actual display is updated only once at the
  // end.

  virtual void new_tuples (Tuple_Queue &tq)
    {
      Tuple **name_number = 0;

      for (Tuple_Queue_Iterator iter (tq);
	   iter.next (name_number) != 0;
	   iter.advance ())
	{
	  if (this->select (*name_number))
	    this->merge_in (*name_number);
	  else
	    ; // ignore Tuple
	}

      this->display ();
    }

  // Finds the place in <tuple_queue_> where the new <Tuple> should be
  // and inserts it there.  Generates the final display representation
  // of the <Tuple> and inserts that in <tuple_string_vector_>.

  void merge_in (Tuple *name_number)
    {
      // Index where the new <Tuple> must be inserted.
      int where;

      for (where = 0;
	   where < tuple_queue_.size ();
	   where++)
	{
	  Tuple **tuple = 0;

	  // Locate the <tuple *> at location <where>.
	  this->tuple_queue_.get (tuple, where);

	  // If <tuple> is less than <name_number> then we've found
	  // where <name_number> belongs so we can exit the loop.
	  if (this->less_than (*tuple, name_number) == 0)
	    break;
	}

      // Now <where> indicates the target index where <name_number>
      // belongs (in sorted order).  Move all other elements one
      // position towards the end in both <tuple_queue_> and
      // <tuple_string_vector_>.

      int size = tuple_queue_.size ();

      for (int k = size - 1;  // Start at end.
	   k >= where;
	   k--)
	{
	  Tuple **t = 0;
	  tuple_queue_.get (t, k);
	  tuple_queue_.set (*t, k + 1);
	      
	  String *s = 0;
	  tuple_string_vector_.get (s, k);
	  tuple_string_vector_.set (*s, k + 1);
	}

      // Now insert new <Tuple> into <tuple_queue_> and its formatted
      // counterpart into <tuple_string_vector_> at position <where>.

      tuple_queue_.set (name_number, where);
      tuple_string_vector_.set (this->format (name_number), where);
    }

  // Displays the contents of <tuple_string_vector_> on the output
  // stream.

  void display (void)
    {
      if (this->tuple_string_vector_.size () > 0)
	{
	  cout << "===========" << endl
	       << "Displaying " << this->display_name_ << endl;

	  String *s = 0;
	  size_t count = 0;

	  for (Tuple_String_Vector_Iterator iter (this->tuple_string_vector_);
	       iter.next (s) != 0;
	       iter.advance ())
	    cout << count++ << ": " << *s;

	  cout << "===========" << endl << endl;
	}
    }

private:
  // <tuple_queue_> contains all <Tuples> that are to be displayed
  // (after filtering!), their order represents the presentation
  // order.
  Tuple_Queue tuple_queue_;       

  // <tuple_string_vector_> parallels <tuple_queue_> and contains the
  // final display presentation of the <Tuples>.  The contents of
  // <tuple_string_vector_> are written into the output stream as is.
  Tuple_String_Vector tuple_string_vector_; 
};

// Main program.
// #M    *** DESIGN PATTERN: ***
// #M    The two Tuple_Displays are registered as observers at the Tupleset.

int 
main (void)
{
  Tupleset tuple_set;

  Tuple_Display *disp1 = 
    new Name_Number_Tuple_Display_1 (String ("chronologically"));

  tuple_set.new_display (disp1);

  Name_Number_Tuple nnt;

  // Runs the event loop that gets the input tuple from the input
  // stream and updates the display.
  nnt.get_tuple (tuple_set, cin);

  return 0;
}
